/*
 * IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING. By
 * downloading, copying, installing or using the software you agree to
 * this license. If you do not agree to this license, do not download,
 * install, copy or use the software.
 * 
 * Intel Open Source License 
 * 
 * Copyright (c) 2004 Intel Corporation. All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *   Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * 
 *   Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * 
 *   Neither the name of the Intel Corporation nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *  
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE INTEL OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef _OASYS_OPTIONS_H_
#define _OASYS_OPTIONS_H_

#include <config.h>
#include <string>
#include <vector>
#include "../compat/inet_aton.h"

namespace oasys {

/**
 * Base class for options. These can be used either with the Getopt
 * class for parsing argv-style declarations or with the OptParser
 * class for parsing argument strings or arrays of strings.
 */
class Opt {
    friend class Getopt;
    friend class OptParser;

public:
    virtual ~Opt();
    
protected:
    /**
     * Private constructor.
     */
    Opt(char shortopt, const char* longopt,
        void* valp, bool* setp, bool needval,
        const char* valdesc, const char* desc);

    /**
     * Virtual callback to set the option to the given string value.
     */
    virtual int set(const char* val, size_t len) = 0;
    
    char shortopt_;
    const char* longopt_;
    void* valp_;
    bool* setp_;
    bool needval_;
    const char* valdesc_;
    const char* desc_;
    Opt*  next_;
};

/**
 * Boolean option class.
 */
class BoolOpt : public Opt {
public:
    /**
     * Basic constructor.
     *
     * @param opt   the option string
     * @param valp  pointer to the value
     * @param desc  descriptive string
     * @param setp  optional pointer to indicate whether or not
                    the option was set
     */
    BoolOpt(const char* opt, bool* valp,
            const char* desc = "", bool* setp = NULL);

    /**
     * Alternative constructor with both short and long options,
     * suitable for getopt calls.
     *
     * @param shortopt  short option character
     * @param longopt   long option string
     * @param valp      pointer to the value
     * @param desc      descriptive string
     * @param setp      optional pointer to indicate whether or not
                        the option was set
     */
    BoolOpt(char shortopt, const char* longopt, bool* valp,
            const char* desc = "", bool* setp = NULL);

protected:
    int set(const char* val, size_t len);
};

/**
 * Integer option class.
 */
class IntOpt : public Opt {
public:
    /**
     * Basic constructor.
     *
     * @param opt     the option string
     * @param valp    pointer to the value
     * @param valdesc short description for the value 
     * @param desc    descriptive string
     * @param setp    optional pointer to indicate whether or not
                      the option was set
     */
    IntOpt(const char* opt, int* valp,
           const char* valdesc = "", const char* desc = "",
           bool* setp = NULL);
    
    /**
     * Alternative constructor with both short and long options,
     * suitable for getopt calls.
     *
     * @param shortopt  short option character
     * @param longopt   long option string
     * @param valp      pointer to the value
     * @param valdesc short description for the value 
     * @param desc      descriptive string
     * @param setp      optional pointer to indicate whether or not
                        the option was set
     */
    IntOpt(char shortopt, const char* longopt, int* valp,
           const char* valdesc = "", const char* desc = "",
           bool* setp = NULL);
    
protected:
    int set(const char* val, size_t len);
};

/**
 * Unsigned integer option class.
 */
class UIntOpt : public Opt {
public:
    /**
     * Basic constructor.
     *
     * @param opt     the option string
     * @param valp    pointer to the value
     * @param valdesc short description for the value 
     * @param desc    descriptive string
     * @param setp    optional pointer to indicate whether or not
                      the option was set
     */
    UIntOpt(const char* opt, u_int* valp,
            const char* valdesc = "", const char* desc = "",
            bool* setp = NULL);
    
    /**
     * Alternative constructor with both short and long options,
     * suitable for getopt calls.
     *
     * @param shortopt  short option character
     * @param longopt   long option string
     * @param valp      pointer to the value
     * @param valdesc short description for the value 
     * @param desc      descriptive string
     * @param setp      optional pointer to indicate whether or not
                        the option was set
     */
    UIntOpt(char shortopt, const char* longopt, u_int* valp,
            const char* valdesc = "", const char* desc = "",
            bool* setp = NULL);
    
protected:
    int set(const char* val, size_t len);
};

/**
 * Unsigned short integer option class.
 */
class UInt16Opt : public Opt {
public:
    /**
     * Basic constructor.
     *
     * @param opt     the option string
     * @param valp    pointer to the value
     * @param valdesc short description for the value 
     * @param desc    descriptive string
     * @param setp    optional pointer to indicate whether or not
                      the option was set
     */
    UInt16Opt(const char* opt, u_int16_t* valp,
              const char* valdesc = "", const char* desc = "",
              bool* setp = NULL);
    
    /**
     * Alternative constructor with both short and long options,
     * suitable for getopt calls.
     *
     * @param shortopt  short option character
     * @param longopt   long option string
     * @param valp      pointer to the value
     * @param valdesc short description for the value 
     * @param desc      descriptive string
     * @param setp      optional pointer to indicate whether or not
                        the option was set
     */
    UInt16Opt(char shortopt, const char* longopt, u_int16_t* valp,
              const char* valdesc = "", const char* desc = "",
              bool* setp = NULL);
    
protected:
    int set(const char* val, size_t len);
};

/**
 * Unsigned byte option class.
 */
class UInt8Opt : public Opt {
public:
   /**
    * Basic constructor.
    *
    * @param opt     the option string
    * @param valp    pointer to the value
    * @param valdesc short description for the value
    * @param desc    descriptive string
    * @param setp    optional pointer to indicate whether or not
                     the option was set
    */
    UInt8Opt(const char* opt, u_int8_t* valp,
             const char* valdesc = "", const char* desc = "",
             bool* setp = NULL);

    /**
     * Alternative constructor with both short and long options,
     * suitable for getopt calls.
     *
     * @param shortopt  short option character
     * @param longopt   long option string
     * @param valp      pointer to the value
     * @param valdesc   short description for the value
     * @param desc      descriptive string
     * @param setp      optional pointer to indicate whether or not
                        the option was set
     */
    UInt8Opt(char shortopt, const char* longopt, u_int8_t* valp,
             const char* valdesc = "", const char* desc = "",
             bool* setp = NULL);

    protected:
        int set(const char* val, size_t len);
};

/**
 * Double option class.
 */
class DoubleOpt : public Opt {
public:
    /**
     * Basic constructor.
     *
     * @param opt     the option string
     * @param valp    pointer to the value
     * @param valdesc short description for the value 
     * @param desc    descriptive string
     * @param setp    optional pointer to indicate whether or not
                      the option was set
     */
    DoubleOpt(const char* opt, double* valp,
              const char* valdesc = "", const char* desc = "",
              bool* setp = NULL);
    
    /**
     * Alternative constructor with both short and long options,
     * suitable for getopt calls.
     *
     * @param shortopt  short option character
     * @param longopt   long option string
     * @param valp      pointer to the value
     * @param valdesc short description for the value 
     * @param desc      descriptive string
     * @param setp      optional pointer to indicate whether or not
                        the option was set
     */
    DoubleOpt(char shortopt, const char* longopt, double* valp,
              const char* valdesc = "", const char* desc = "",
              bool* setp = NULL);
    
protected:
    int set(const char* val, size_t len);
};

/**
 * String option class.
 */
class StringOpt : public Opt {
public:
    /**
     * Basic constructor.
     *
     * @param opt     the option string
     * @param valp    pointer to the value
     * @param valdesc short description for the value 
     * @param desc    descriptive string
     * @param setp    optional pointer to indicate whether or not
                      the option was set
     */
    StringOpt(const char* opt, std::string* valp,
              const char* valdesc = "", const char* desc = "",
              bool* setp = NULL);
    
    /**
     * Alternative constructor with both short and long options,
     * suitable for getopt calls.
     *
     * @param shortopt  short option character
     * @param longopt   long option string
     * @param valp      pointer to the value
     * @param valdesc	short description for the value 
     * @param desc      descriptive string
     * @param setp      optional pointer to indicate whether or not
                        the option was set
     */
    StringOpt(char shortopt, const char* longopt, std::string* valp,
              const char* valdesc = "", const char* desc = "",
              bool* setp = NULL);

protected:
    int set(const char* val, size_t len);
};

/**
 * Char buffer option class.
 */
class CharBufOpt : public Opt {
public:
    /**
     * Basic constructor.
     *
     * @param opt     the option string
     * @param valp    pointer to the value
     * @param lenp    pointer to the length
     * @param buflen  length of the buffer
     * @param valdesc short description for the value 
     * @param desc    descriptive string
     * @param setp    optional pointer to indicate whether or not
                      the option was set
     */
    CharBufOpt(const char* opt, char* valp, size_t* lenp, size_t buflen,
               const char* valdesc = "", const char* desc = "",
               bool* setp = NULL);
    
    /**
     * Alternative constructor with both short and long options,
     * suitable for getopt calls.
     *
     * @param shortopt  short option character
     * @param longopt   long option string
     * @param valp      pointer to the value
     * @param lenp    pointer to the length
     * @param buflen  length of the buffer
     * @param valdesc	short description for the value 
     * @param desc      descriptive string
     * @param setp      optional pointer to indicate whether or not
                        the option was set
     */
    CharBufOpt(char shortopt, const char* longopt,
               char* valp, size_t* lenp, size_t buflen,
               const char* valdesc = "", const char* desc = "",
               bool* setp = NULL);
    
protected:
    size_t buflen_;
    size_t* lenp_;
    
    int set(const char* val, size_t len);
};

/**
 * Internet address (dotted-quad or DNS name) option class.
 */
class InAddrOpt : public Opt {
public:
    /**
     * Basic constructor.
     *
     * @param opt     the option string
     * @param valp    pointer to the value
     * @param valdesc short description for the value 
     * @param desc    descriptive string
     * @param setp    optional pointer to indicate whether or not
                      the option was set
     */
    InAddrOpt(const char* opt, in_addr_t* valp,
              const char* valdesc = "", const char* desc = "",
              bool* setp = NULL);
    
    /**
     * Alternative constructor with both short and long options,
     * suitable for getopt calls.
     *
     * @param shortopt  short option character
     * @param longopt   long option string
     * @param valp      pointer to the value
     * @param valdesc 	short description for the value 
     * @param desc      descriptive string
     * @param setp      optional pointer to indicate whether or not
                        the option was set
     */
    InAddrOpt(char shortopt, const char* longopt, in_addr_t* valp,
              const char* valdesc = "", const char* desc = "",
              bool* setp = NULL);

protected:
    int set(const char* val, size_t len);
};

/**
 * Option class to select one of a set of potential values based on
 * string keys.
 */
class EnumOpt : public Opt {
public:
    struct Case {
        const char* key;
        int         val;
    };
    
    /**
     * Basic constructor.
     *
     * @param opt     option string
     * @param cases   pointer to the array of cases
     * @param valp    pointer to the value
     * @param valdesc short description for the value 
     * @param desc    descriptive string
     * @param setp    optional pointer to indicate whether or not
                      the option was set
     */
    EnumOpt(const char* opt, Case* cases, int* valp,
            const char* valdesc = "", const char* desc = "",
            bool* setp = NULL);
    
    /**
     * Alternative constructor with both short and long options,
     * suitable for getopt calls.
     *
     * @param shortopt  short option character
     * @param longopt   long option string
     * @param cases     pointer to the array of cases
     * @param valp      pointer to the value
     * @param valdesc 	short description for the value 
     * @param desc      descriptive string
     * @param setp      optional pointer to indicate whether or not
                        the option was set
     */
    EnumOpt(char shortopt, const char* longopt,
            Case* cases, int* valp,
            const char* valdesc = "", const char* desc = "",
            bool* setp = NULL);

protected:
    int set(const char* val, size_t len);
    Case* cases_;
};

#ifdef OASYS_BLUETOOTH_ENABLED
#include <bluetooth/bluetooth.h>
/**
 * Bluetooth address (colon-separated hex) option class.
 */
class BdAddrOpt : public Opt {
public:
   /**
    * Basic constructor.
    *
    * @param opt     the option string
    * @param valp    pointer to the value
    * @param valdesc short description for the value
    * @param desc    descriptive string
    * @param setp    optional pointer to indicate whether or not
                     the option was set
    */
    BdAddrOpt(const char* opt, bdaddr_t* valp,
              const char* valdesc = "", const char* desc = "",
              bool* setp = NULL);

   /**
    * Alternative constructor with both short and long options,
    * suitable for getopt calls.
    *
    * @param shortopt  short option character
    * @param longopt   long option string
    * @param valp      pointer to the value
    * @param valdesc   short description for the value
    * @param desc      descriptive string
    * @param setp      optional pointer to indicate whether or not 
                       the option was set
    */
    BdAddrOpt(char shortopt, const char* longopt, bdaddr_t* valp,
              const char* valdesc = "", const char* desc = "",
              bool* setp = NULL);

protected:
    int set(const char* val, size_t len);
};
#endif // OASYS_BLUETOOTH_ENABLED

} // namespace oasys

#endif /* _OASYS_OPTIONS_H_ */
