/*
 * IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING. By
 * downloading, copying, installing or using the software you agree to
 * this license. If you do not agree to this license, do not download,
 * install, copy or use the software.
 * 
 * Intel Open Source License 
 * 
 * Copyright (c) 2004 Intel Corporation. All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *   Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * 
 *   Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * 
 *   Neither the name of the Intel Corporation nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *  
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE INTEL OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "URL.h"
#include "StringBuffer.h"
#include "debug/DebugUtils.h"

namespace oasys {

urlerr_t
URL::parse_internal()
{
    size_t beg, end;

    // reset all fields
    proto_.erase();
    host_.erase();
    port_ = 0;
    path_.erase();

    // extract protocol
    if ((end = url_.find(':')) == std::string::npos)
        return URLPARSE_BADSEP;
    
    // skip the separator
    if (url_[end+1] != '/' || url_[end+2] != '/')
        return URLPARSE_BADPROTO;

    // valid protocol -- assign it and skip
    proto_.assign(url_, 0, end);
    beg = end + 3;

    // check for no host
    if (beg == url_.length())
        return URLPARSE_NOHOST;

    // search for the hostname/path separator
    end = url_.find_first_of(":/", beg, 2);

    if (end == std::string::npos) {
        // there's no port or pathname, it's just a host name, but
        // append a / for the canonical form
        host_.assign(url_, beg, url_.length());
        url_.push_back('/');
        return URLPARSE_OK;
    }

    // grab the hostname
    host_.assign(url_, beg, end - beg);

    // parse the port if it exists
    if (url_[end] == ':') {
        char* endstr;
        beg = end + 1;
        end = url_.find('/', beg);
        if (end == std::string::npos) {
            end = url_.length();

            // again, adjust for no trailing slash
            url_.push_back('/');
        }
        
        std::string portstr(url_, beg, end - beg);
        u_int32_t portval = strtoul(portstr.c_str(), &endstr, 10);
        
        if (portval > 65535)
            return URLPARSE_BADPORT;

        if (endstr != (portstr.c_str() + portstr.length()))
            return URLPARSE_BADPORT;

        port_ = portval;

        beg = end + 1;
    } else {
        // otherwise just skip past the slash separator
        ASSERT(url_[end] == '/');
        beg = end + 1;
    }

    // check for an empty path
    if (beg >= url_.length())
        return URLPARSE_OK;

    // otherwise store the path and we're done
    path_.assign(url_, beg, url_.length() - beg);
    return URLPARSE_OK;
}

urlerr_t
URL::parse()
{
    err_ = parse_internal();
    return err_;
}

void
URL::format(const std::string& proto, const std::string& host, u_int16_t port,
            const std::string& path)
{
    StringBuffer buf;
    
    proto_ = proto;
    host_ = host;
    port_ = port;
    path_ = path;

    buf.append(proto);
    buf.append("://");
    buf.append(host);

    if (port != 0) {
        buf.appendf(":%d", port);
    }

    if (path[0] != '/')
        buf.append('/');

    buf.append(path);
    
    url_.assign(buf.data(), buf.length());
    err_ = URLPARSE_OK;
}

} // namespace oasys
